import pytest
import os
import json
import uuid
import time
import urllib3
from tenacity import retry, stop_after_attempt, wait_fixed, retry_if_exception_type

from psc_access_integration.api_clients.adminservices import AdminServicesApiAccess
from psc_auth_integration.entities.auth_user_session import AuthUserSession
from assignments.common.utils import DEFENSE_URL, SENSOR_GROUPS_UPGRADE_USERNAME, SENSOR_GROUPS_UPGRADE_PASSWORD, ENV, sim_endpoint, \
    create_asset_group, \
    add_policy_override, add_asset_group_to_endpoint, LOCAL_PAS_HOST, EXPECTED_CALL_TO_SUCCEED, execute, \
    delete_sensor, delete_policy, delete_asset_group, remove_asset_group_from_endpoint, CSR_HOST, delete_asset_group_with_ids, \
    delete_sensors, delete_policies_with_ids
from policy_service_integration.api_client.policyservice import PolicyApi
from python_sensor_sim.enums import SensorDeviceType
from assignments.tests.test_policy_preview import search_assets

urllib3.disable_warnings()


SESSION = None
CREATED_ASSET_GROUPS = None
CREATED_POLICIES = None
CREATED_ENDPOINTS = None


@pytest.mark.validate
def test_sensor_groups_upgrade():
    try:
        global CREATED_ASSET_GROUPS, CREATED_POLICIES, CREATED_ENDPOINTS, SESSION
        os.environ["ENV"] = ENV
        SESSION = AuthUserSession(gateway_url=DEFENSE_URL, username=SENSOR_GROUPS_UPGRADE_USERNAME, password=SENSOR_GROUPS_UPGRADE_PASSWORD)

        device_ids = create_sensors(18)
        CREATED_ENDPOINTS = device_ids
        policy_ids = create_policies(5)
        CREATED_POLICIES = policy_ids[3:]

        time.sleep(120)

        search_response_assets_ids = search_assets(SESSION, search_definition={})
        assert len(device_ids) == len(search_response_assets_ids)

        search_response_assets_ids = search_assets(SESSION, query="_exists_:policy_assignment_type")
        assert 0 == len(search_response_assets_ids)

        update_feature_flag({ "assetGroupsManagement": True })

        sensor_groups_upgrade()

        time.sleep(120)

        search_response_assets_ids = search_assets(SESSION, query="_exists_:policy_assignment_type")
        assert len(device_ids) == len(search_response_assets_ids)

        search_response_assets_ids = search_assets(SESSION, query="policy_assignment_type: DEFAULT")
        assert len(device_ids) == len(search_response_assets_ids)

        asset_ids_to_override = [ device_ids[-1].get("device_id"), device_ids[-2].get("device_id"), device_ids[-3].get("device_id") ]
        for asset_id in asset_ids_to_override:
            add_policy_override(session=SESSION, policy_id=policy_ids[-1], device_id=asset_id)

        time.sleep(60)

        search_response_assets_ids = search_assets(SESSION, query="policy_assignment_type: MANUAL")
        assert len(asset_ids_to_override) == len(search_response_assets_ids)
        assert set(asset_ids_to_override) == set(search_response_assets_ids)

        asset_group_ids = create_asset_groups(policy_ids)
        CREATED_ASSET_GROUPS = asset_group_ids

        asset_ids_to_manually_assign_to_group_1 = [ device_ids[0].get("device_id"), device_ids[1].get("device_id"), device_ids[2].get("device_id")]
        asset_ids_to_manually_assign_to_group_2 = [ device_ids[3].get("device_id"), device_ids[4].get("device_id"), device_ids[5].get("device_id")]
        asset_ids_to_manually_assign_to_group_3 = [ device_ids[6].get("device_id"), device_ids[7].get("device_id"), device_ids[8].get("device_id")]
        asset_ids_to_manually_assign_to_group_4 = [ device_ids[9].get("device_id"), device_ids[10].get("device_id"), device_ids[11].get("device_id")]
        asset_ids_to_manually_assign_to_group_5 = [ device_ids[12].get("device_id"), device_ids[13].get("device_id"), device_ids[14].get("device_id")]

        add_asset_group_to_endpoint(session=SESSION, asset_group_id=asset_group_ids[0], endpoint_ids=asset_ids_to_manually_assign_to_group_1)
        add_asset_group_to_endpoint(session=SESSION, asset_group_id=asset_group_ids[1], endpoint_ids=asset_ids_to_manually_assign_to_group_2)
        add_asset_group_to_endpoint(session=SESSION, asset_group_id=asset_group_ids[2], endpoint_ids=asset_ids_to_manually_assign_to_group_3)
        add_asset_group_to_endpoint(session=SESSION, asset_group_id=asset_group_ids[3], endpoint_ids=asset_ids_to_manually_assign_to_group_4)
        add_asset_group_to_endpoint(session=SESSION, asset_group_id=asset_group_ids[4], endpoint_ids=asset_ids_to_manually_assign_to_group_5)

        asset_ids_to_reevaluate = []
        for group_manual_assignments_asset_ids in [asset_ids_to_manually_assign_to_group_1, asset_ids_to_manually_assign_to_group_2, asset_ids_to_manually_assign_to_group_3, asset_ids_to_manually_assign_to_group_4, asset_ids_to_manually_assign_to_group_5]:
            for asset_id in group_manual_assignments_asset_ids:
                asset_ids_to_reevaluate.append(asset_id)

        reevaluate(asset_ids_to_reevaluate)
        time.sleep(120)

        reevaluate_processed_ids = [ id for id in asset_ids_to_reevaluate if id not in asset_ids_to_manually_assign_to_group_5 ]

        search_response_assets_ids = search_assets(SESSION, query="policy_assignment_type: GROUP")
        assert len(reevaluate_processed_ids) == len(search_response_assets_ids)
        assert set(reevaluate_processed_ids) == set(search_response_assets_ids)

        query=f'(id: {device_ids[12].get("device_id")} OR {device_ids[13].get("device_id")} OR {device_ids[14].get("device_id")}) AND policy_assignment_type: DEFAULT'
        search_response_assets_ids = search_assets(SESSION, query=query)
        assert len(asset_ids_to_manually_assign_to_group_5) == len(search_response_assets_ids)
        assert set(asset_ids_to_manually_assign_to_group_5) == set(search_response_assets_ids)

        remove_asset_group_from_endpoint(SESSION, asset_group_ids[0], asset_ids_to_manually_assign_to_group_1)
        time.sleep(120)
        reevaluate(asset_ids_to_manually_assign_to_group_1)
        time.sleep(120)

        assets_with_default_type = [ device_ids[12].get("device_id"), device_ids[13].get("device_id"), device_ids[14].get("device_id") ]
        assets_with_default_type.extend(asset_ids_to_manually_assign_to_group_1)

        search_response_assets_ids = search_assets(SESSION, query="policy_assignment_type: DEFAULT")
        assert len(assets_with_default_type) == len(search_response_assets_ids)
        assert set(assets_with_default_type) == set(search_response_assets_ids)
    except Exception as e:
        print(f"Exception: {e}")
        assert False
    finally:
        if SESSION:
            if CREATED_ASSET_GROUPS is not None:
                delete_asset_group_with_ids(SESSION, CREATED_ASSET_GROUPS)

            if CREATED_ENDPOINTS is not None:
                delete_sensors(SESSION, CREATED_ENDPOINTS)

            if CREATED_POLICIES is not None:
                delete_policies_with_ids(SESSION, CREATED_POLICIES)

            update_feature_flag({ "assetGroupsManagement": False })


def create_policies(count):
    policy_api = PolicyApi(SESSION)
    response = policy_api.get_policies_summary(org_key=SESSION.get_user_org_key())
    assert response.ok, EXPECTED_CALL_TO_SUCCEED
    policies = json.loads(response.content).get("policies")

    response = policy_api.get_policy_details(org_key=SESSION.get_user_org_key(), policy_id=policies[0].get("id"))
    assert response.ok, EXPECTED_CALL_TO_SUCCEED
    policy_details = json.loads(response.content)

    created_policy_ids = [ policy.get("id") for policy in policies ]
    for _ in range(1, count + 1, 1):
        policy_details['name'] = str(uuid.uuid4())
        policy_details['id'] = None
        policy_details['description'] = f"{policy_details['name']} dummy description."

        create_policy_response = policy_api.create_policy(org_key=SESSION.get_user_org_key(), body=policy_details)
        assert create_policy_response.ok, EXPECTED_CALL_TO_SUCCEED

        created_policy_ids.append(json.loads(create_policy_response.content).get('id'))

    return created_policy_ids


def create_sensors(count):
    devices = []
    for i in range(1, count + 1, 1):
        device = sim_endpoint(SESSION, suffix=f"upgrade-{i}", device_type=SensorDeviceType.WINDOWS)
        devices.append(device)

    return devices


def update_feature_flag(payload):
    response = SESSION.patch(
        url=f"{CSR_HOST}/adminServices/v5/orgs/{SESSION.get_user_org_id()}/settings/ui-feature-flags",
        data=json.dumps(payload)
    )
    assert response.ok, EXPECTED_CALL_TO_SUCCEED


def create_asset_groups(policy_ids):
    group_1 = create_asset_group(SESSION, name=f"upgrade-{str(uuid.uuid4())}", policy_id=policy_ids[3])
    group_2 = create_asset_group(SESSION, name=f"upgrade-{str(uuid.uuid4())}", policy_id=policy_ids[3])
    group_3 = create_asset_group(SESSION, name=f"upgrade-{str(uuid.uuid4())}", policy_id=policy_ids[4])
    group_4 = create_asset_group(SESSION, name=f"upgrade-{str(uuid.uuid4())}", policy_id=policy_ids[5])
    group_5 = create_asset_group(SESSION, name=f"upgrade-{str(uuid.uuid4())}")

    group_ids = []
    for group in [ group_1, group_2, group_3, group_4, group_5 ]:
        group_ids.append(group.get("id"))

    return group_ids


@retry(retry=retry_if_exception_type(), reraise=True, stop=stop_after_attempt(20), wait=wait_fixed(5))
def sensor_groups_upgrade():
    response = SESSION.post(
        url=f"{LOCAL_PAS_HOST}/private/orgs/{SESSION.get_user_org_key()}/upgrade_to_asset_groups",
        data=json.dumps({})
    )
    assert response.ok, EXPECTED_CALL_TO_SUCCEED


@retry(retry=retry_if_exception_type(), reraise=True, stop=stop_after_attempt(20), wait=wait_fixed(5))
def reevaluate(asset_ids):
    response = SESSION.post(
        url=f"{LOCAL_PAS_HOST}/private/orgs/{SESSION.get_user_org_key()}/reevaluate",
        data=json.dumps({
            "asset_ids": asset_ids
        })
    )
    assert response.ok, EXPECTED_CALL_TO_SUCCEED
